@@ -19,6 +19,7 @@ import urllib
 import requests
 import six
 import urllib3
+import time
 from requests import exceptions as req_exc
 
 try:
@@ -49,16 +50,58 @@ def reraise_neutron_exception(func):
 
     @functools.wraps(func)
     def callee(*args, **kwargs):
+        """Catches third-party exceptions and raises Infoblox exceptions"""
         try:
             return func(*args, **kwargs)
         except req_exc.Timeout as e:
-            raise ib_ex.InfobloxTimeoutError(e)
+            raise ib_ex.InfobloxTimeoutError(e, reason=e)
         except req_exc.RequestException as e:
             raise ib_ex.InfobloxConnectionError(reason=e)
 
     return callee
 
 
+def retry_on_expired_cookie(func):
+    """
+    Decorator to handle expired cookies by re-authenticating
+    and retrying the request.
+    """
+    @functools.wraps(func)
+    def wrapper(self, *args, **kwargs):
+        """
+        Wrapper function to handle expired cookies by re-authenticating
+        and retrying the request.
+        """
+        try:
+            return func(self, *args, **kwargs)
+        except (req_exc.HTTPError, ib_ex.InfobloxBadWAPICredential,
+                ib_ex.InfobloxFileDownloadFailed,
+                ib_ex.InfobloxFileUploadFailed) as e:
+            if isinstance(e, (req_exc.HTTPError,
+                              ib_ex.InfobloxBadWAPICredential,
+                              ib_ex.InfobloxFileDownloadFailed,
+                              ib_ex.InfobloxFileUploadFailed)):
+                LOG.warning("Bad WAPI credentials or Cookie timeout.\
+                         Re-authenticating and retrying the request.")
+            else:
+                raise
+
+            LOG.warning("Clearing cookies and re-authenticating.")
+            self.session.cookies.clear()
+            if hasattr(self, 'username') and hasattr(self, 'password'):
+                LOG.warning("Re-authenticating with username and password.")
+                self.session.auth = (self.username, self.password)
+            elif hasattr(self, 'cert') and hasattr(self, 'key'):
+                LOG.warning("Re-authenticating with client certificate.")
+                self.session.cert = (self.cert, self.key)
+            else:
+                LOG.warning("No valid credentials found to re-authenticate.")
+                raise ib_ex.InfobloxConfigException(
+                    msg="No valid credentials found to re-authenticate.")
+            return func(self, *args, **kwargs)
+    return wrapper
+
+
 class Connector(object):
     """Connector stands for interacting with Infoblox NIOS
 
@@ -90,9 +133,12 @@ class Connector(object):
             self._urlencode = urlparse.urlencode
             self._quote = urlparse.quote
             self._urljoin = urlparse.urljoin
+        masked_options = utils.mask_sensitive_data(options)
+        LOG.debug(
+            "Connector initialized with options: {}".format(masked_options))
 
     def _parse_options(self, options):
-        """Copy needed options to self"""
+        """Copy necessary options to self"""
         attributes = ('host', 'wapi_version', 'username', 'password',
                       'cert', 'key', 'ssl_verify', 'http_request_timeout',
                       'max_retries', 'http_pool_connections',
@@ -116,6 +162,7 @@ class Connector(object):
                 setattr(self, attr, self.DEFAULT_OPTIONS[attr])
             elif attr not in creds:
                 msg = "WAPI config error. Option %s is not defined" % attr
+                LOG.error(msg)
                 raise ib_ex.InfobloxConfigException(msg=msg)
 
         def check_creds(credentials):
@@ -131,6 +178,7 @@ class Connector(object):
                 not check_creds(creds_cert_auth):
             msg = "WAPI config error. Option either (username, password) " \
                   "or (cert, key) should be passed"
+            LOG.error(msg)
             raise ib_ex.InfobloxConfigException(msg=msg)
 
         self.wapi_url = "https://%s/wapi/v%s/" % (self.host,
@@ -139,6 +187,7 @@ class Connector(object):
             self.wapi_version)
 
     def _configure_session(self):
+        LOG.debug("Configuring session")
         self.session = requests.Session()
         adapter = requests.adapters.HTTPAdapter(
             pool_connections=self.http_pool_connections,
@@ -147,22 +196,31 @@ class Connector(object):
         self.session.mount('http://', adapter)
         self.session.mount('https://', adapter)
         if hasattr(self, 'username') and hasattr(self, 'password'):
+            LOG.info("Authenticating with username and password.")
             self.session.auth = (self.username, self.password)
         else:
             self.session.cert = (self.cert, self.key)
+            LOG.info("Authenticating with client certificate.")
         self.session.verify = utils.try_value_to_bool(self.ssl_verify,
                                                       strict_mode=False)
+        LOG.debug("SSL verification is %s", self.session.verify)
 
         if self.silent_ssl_warnings:
+            LOG.debug("Silencing SSL warnings")
             urllib3.disable_warnings()
 
     def _construct_url(self, relative_path, query_params=None,
                        extattrs=None, force_proxy=False):
+        """
+        Construct URL for Infoblox WAPI request
+        """
+        LOG.debug("Constructing URL for relative path: %s", relative_path)
         if query_params is None:
             query_params = {}
         if extattrs is None:
             extattrs = {}
         if force_proxy:
+            LOG.debug("Forcing proxy search")
             query_params['_proxy_search'] = 'GM'
 
         if not relative_path or relative_path[0] == '/':
@@ -189,6 +247,7 @@ class Connector(object):
 
         base_url = self._urljoin(self.wapi_url,
                                  self._quote(relative_path))
+        LOG.debug("Constructed URL: %s", base_url + query)
         return base_url + query
 
     @staticmethod
@@ -202,12 +261,18 @@ class Connector(object):
         if not obj_type:
             raise ValueError('NIOS object type cannot be empty.')
         if obj_type_expected and '/' in obj_type:
+            LOG.error(
+                "NIOS object type cannot contain slash: {}".format(obj_type))
             raise ValueError('NIOS object type cannot contain slash.')
 
     @staticmethod
     def _validate_authorized(response):
         if response.status_code == requests.codes.UNAUTHORIZED:
-            raise ib_ex.InfobloxBadWAPICredential(response='')
+            raise ib_ex.InfobloxBadWAPICredential(
+                response=response.content,
+                content=utils.format_html(response.content),
+                code=response.status_code
+            )
 
     @staticmethod
     def _build_query_params(payload=None, return_fields=None,
@@ -233,6 +298,37 @@ class Connector(object):
 
         return query_params
 
+    def is_cookie_expired(self):
+        """
+        Check if the cookie is expired by comparing the expiration time
+        with the current time.
+        """
+        LOG.debug("Validating cookie expiration")
+        cookie_jar = self.session.cookies
+        for cookie in cookie_jar:
+            if cookie.name == 'ibapauth':
+                # Parse the cookie value
+                cookie_value = cookie.value.strip('"')
+                cookie_parts = {}
+                for part in cookie_value.split(','):
+                    if '=' in part:
+                        key, value = part.split('=', 1)
+                        cookie_parts[key] = value
+
+                # Extract ctime and timeout
+                ctime = int(cookie_parts.get('ctime', 0))
+                timeout = int(cookie_parts.get('timeout', 0))
+
+                # Calculate expiration time
+                expiration_time = ctime + timeout
+
+                # Get current time
+                current_time = int(time.time())
+
+                # Check if the cookie is expired
+                return current_time > expiration_time
+        return True  # If the cookie is not found, consider it expired
+
     def _get_request_options(self, data=None):
         opts = dict(timeout=self.http_request_timeout,
                     headers=self.DEFAULT_HEADER,
@@ -241,6 +337,25 @@ class Connector(object):
             opts['data'] = jsonutils.dumps(data)
         return opts
 
+    def _validate_cookie(self):
+        if self.is_cookie_expired():
+            LOG.info("Cookie expired. \
+            Clearing cookies and re-authenticating on the next request.")
+            self.session.cookies.clear()
+            if hasattr(self, 'username') and hasattr(self, 'password'):
+                LOG.info("Re-authenticating with username and password.")
+                self.session.auth = (self.username, self.password)
+            elif hasattr(self, 'cert') and hasattr(self, 'key'):
+                LOG.info("Re-authenticating with client certificate.")
+                self.session.cert = (self.cert, self.key)
+            else:
+                LOG.warning("No valid credentials found to re-authenticate.")
+                raise ib_ex.InfobloxConfigException(
+                    msg="No valid credentials found to re-authenticate.")
+        else:
+            LOG.info("Using existing cookie for authentication")
+            self.session.auth = None  # Do not re-authenticate
+
     @staticmethod
     def _parse_reply(request):
         """Tries to parse reply from NIOS.
@@ -248,8 +363,10 @@ class Connector(object):
         Raises exception with content if reply is not in json format
         """
         try:
+            LOG.debug("WAPI Response: %s", request.content)
             return jsonutils.loads(request.content)
         except ValueError:
+            LOG.error("Failed to parse reply from NIOS: %s", request.content)
             raise ib_ex.InfobloxConnectionError(reason=request.content)
 
     def _log_request(self, verb, url, opts):
@@ -261,6 +378,7 @@ class Connector(object):
             LOG.debug(*message)
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def get_object(self, obj_type, payload=None, return_fields=None,
                    extattrs=None, force_proxy=False, max_results=None,
                    paging=None):
@@ -269,7 +387,7 @@ class Connector(object):
         Some get requests like 'ipv4address' should be always
         proxied to GM on Hellfire
         If request is cloud and proxy is not forced yet,
-        then plan to do 2 request:
+        then plan to do two requests:
         - the first one is not proxied to GM
         - the second is proxied to GM
 
@@ -282,13 +400,13 @@ class Connector(object):
             force_proxy   (bool): Set _proxy_search flag
                                   to process requests on GM
             max_results   (int): Maximum number of objects to be returned.
-                If set to a negative number the appliance will return an error
+                If set to a negative number, the appliance will return an error
                 when the number of returned objects would exceed the setting.
                 The default is -1000. If this is set to a positive number,
                 the results will be truncated when necessary.
-            paging    (bool): Enables paging to wapi calls if paging = True,
+            Paging (bool): Enables paging to wapi calls if paging = True,
                 it uses _max_results to set paging size of the wapi calls.
-                If _max_results is negative it will take paging size as 1000.
+                If _max_results is negative, it will take paging size as 1000.
 
         Returns:
             A list of the Infoblox objects requested
@@ -314,9 +432,11 @@ class Connector(object):
         proxy_flag = self.cloud_api_enabled and force_proxy
 
         try:
-            return self._handle_get_object(obj_type, query_params, extattrs,
-                                           proxy_flag)
+            return self._handle_get_object(
+                obj_type, query_params, extattrs, proxy_flag)
         except req_exc.HTTPError:
+            LOG.warning(
+                "Failed on object search with proxy flag %s", proxy_flag)
             # Do second get call with force_proxy if not done yet
             return self._handle_get_object(obj_type, query_params,
                                            extattrs, proxy_flag=True)
@@ -359,7 +479,7 @@ class Connector(object):
         if self.session.cookies:
             # the first 'get' or 'post' action will generate a cookie
             # after that, we don't need to re-authenticate
-            self.session.auth = None
+            self._validate_cookie()
         r = self.session.get(url, **opts)
 
         self._validate_authorized(r)
@@ -371,21 +491,35 @@ class Connector(object):
         return self._parse_reply(r)
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def download_file(self, url):
+        LOG.debug("Downloading file from %s", url)
+        req_cookies = None
         if self.session.cookies:
-            self.session.auth = None
+            self._validate_cookie()
+            cookies_dict = self.session.cookies.get_dict()
+            ibapauth_cookie = cookies_dict.get('ibapauth', None)
+            if ibapauth_cookie:
+                req_cookies = {'ibapauth': ibapauth_cookie}
+            else:
+                req_cookies = None
+
         headers = {'content-type': 'application/force-download'}
-        r = self.session.get(url, headers=headers)
+        r = self.session.get(url, headers=headers, cookies=req_cookies)
         if r.status_code != requests.codes.ok:
+            res_content = utils.format_html(r.content)
+            LOG.error("Failed to download file from %s: %s", url, res_content)
             response = utils.safe_json_load(r.content)
             raise ib_ex.InfobloxFileDownloadFailed(
                 response=response,
-                url=url
+                url=url,
+                content=res_content,
+                code=r.status_code
             )
-
         return r
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def upload_file(self, url, files):
         """Upload file to fully-qualified upload url
 
@@ -397,31 +531,41 @@ class Connector(object):
         Raises:
             InfobloxException
         """
+        LOG.debug("Uploading file to %s", url)
+        req_cookies = None
         if self.session.cookies:
-            self.session.auth = None
-        r = self.session.post(url, files=files)
+            self._validate_cookie()
+            cookies_dict = self.session.cookies.get_dict()
+            ibapauth_cookie = cookies_dict.get('ibapauth', None)
+            if ibapauth_cookie:
+                req_cookies = {'ibapauth': ibapauth_cookie}
+            else:
+                req_cookies = None
+        r = self.session.post(url, files=files, cookies=req_cookies)
         if r.status_code != requests.codes.ok:
             response = utils.safe_json_load(r.content)
+            res_content = utils.format_html(r.content)
+            LOG.error("Failed to upload file to %s: %s", url, res_content)
             raise ib_ex.InfobloxFileUploadFailed(
                 response=response,
                 url=url,
-                content=response,
+                content=res_content,
                 code=r.status_code,
             )
-
         return r
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def create_object(self, obj_type, payload, return_fields=None):
         """Create an Infoblox object of type 'obj_type'
 
         Args:
-            obj_type        (str): Infoblox object type,
-                                  e.g. 'network', 'range', etc.
-            payload       (dict): Payload with data to send
+            obj_type (str): Infoblox object type,
+                            e.g. 'network', 'range', etc.
+            payload (dict): Payload with data to send
             return_fields (list): List of fields to be returned
         Returns:
-            The object reference of the newly create object
+            The object reference of the new creation object
         Raises:
             InfobloxException
         """
@@ -435,12 +579,14 @@ class Connector(object):
         if self.session.cookies:
             # the first 'get' or 'post' action will generate a cookie
             # after that, we don't need to re-authenticate
-            self.session.auth = None
+            self._validate_cookie()
         r = self.session.post(url, **opts)
 
         self._validate_authorized(r)
 
         if r.status_code != requests.codes.CREATED:
+            LOG.error(
+                "Failed to create object with url %s: %s", url, r.content)
             response = utils.safe_json_load(r.content)
             already_assigned = 'is assigned to another network view'
             if response and already_assigned in response.get('text'):
@@ -466,7 +612,11 @@ class Connector(object):
                 code=resp.status_code)
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def call_func(self, func_name, ref, payload, return_fields=None):
+        LOG.debug("Calling function %s on object %s", func_name, ref)
+        if self.session.cookies:
+            self._validate_cookie()
         query_params = self._build_query_params(return_fields=return_fields)
         query_params['_function'] = func_name
 
@@ -491,6 +641,7 @@ class Connector(object):
         return self._parse_reply(r)
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def update_object(self, ref, payload, return_fields=None):
         """Update an Infoblox object
 
@@ -508,6 +659,8 @@ class Connector(object):
         opts = self._get_request_options(data=payload)
         url = self._construct_url(ref, query_params)
         self._log_request('put', url, opts)
+        if self.session.cookies:
+            self._validate_cookie()
         r = self.session.put(url, **opts)
 
         self._validate_authorized(r)
@@ -524,6 +677,7 @@ class Connector(object):
         return self._parse_reply(r)
 
     @reraise_neutron_exception
+    @retry_on_expired_cookie
     def delete_object(self, ref, delete_arguments=None):
         """Remove an Infoblox object
 
@@ -540,6 +694,8 @@ class Connector(object):
             delete_arguments = {}
         url = self._construct_url(ref, query_params=delete_arguments)
         self._log_request('delete', url, opts)
+        if self.session.cookies:
+            self._validate_cookie()
         r = self.session.delete(url, **opts)
 
         self._validate_authorized(r)
@@ -575,5 +731,7 @@ class Connector(object):
         version_match = re.search(r'(\d+)\.(\d+)', wapi_version)
         if version_match:
             if int(version_match.group(1)) >= CLOUD_WAPI_MAJOR_VERSION:
+                LOG.debug("Cloud WAPI version detected: %s", wapi_version)
                 return True
+        LOG.debug("Non-cloud WAPI version detected: %s", wapi_version)
         return False
